local GearEvaluator = _G.GearEvaluator
local GearEvaluator_Searcher = GearEvaluator:NewModule("Searcher", "AceTimer-3.0", "AceEvent-3.0", "AceBucket-3.0")

GearEvaluator_Searcher.slotNames = {
  "HeadSlot",
  "NeckSlot",
  "ShoulderSlot",
  "BackSlot",
  "ChestSlot",
  "ShirtSlot",
  "TabardSlot",
  "WristSlot",
  "HandsSlot",
  "WaistSlot",
  "LegsSlot",
  "FeetSlot",
  "Finger0Slot",
  "Finger1Slot",
  "Trinket0Slot",
  "Trinket1Slot",
  "MainHandSlot",
  "SecondaryHandSlot",
  "RangedSlot",
  "AmmoSlot",
}

GearEvaluator_Searcher.invSlots = {
  ["INVTYPE_AMMO"] = "AmmoSlot",
  ["INVTYPE_HEAD"] = "HeadSlot",
  ["INVTYPE_NECK"] = "NeckSlot",
  ["INVTYPE_SHOULDER"] = "ShoulderSlot",
  ["INVTYPE_BODY"] = "ShirtSlot",
  ["INVTYPE_CHEST"] = "ChestSlot",
  ["INVTYPE_ROBE"] = "ChestSlot",
  ["INVTYPE_WAIST"] = "WaistSlot",
  ["INVTYPE_LEGS"] = "LegsSlot",
  ["INVTYPE_FEET"] = "FeetSlot",
  ["INVTYPE_WRIST"] = "WristSlot",
  ["INVTYPE_HAND"] = "HandsSlot",
  ["INVTYPE_FINGER"] = "Finger0Slot",
  ["INVTYPE_TRINKET"] = "Trinket0Slot",
  ["INVTYPE_CLOAK"] = "BackSlot",
  ["INVTYPE_WEAPON"] = "MainHandSlot",
  ["INVTYPE_SHIELD"] = "SecondaryHandSlot",
  ["INVTYPE_2HWEAPON"] = "MainHandSlot",
  ["INVTYPE_WEAPONMAINHAND"] = "MainHandSlot",
  ["INVTYPE_WEAPONOFFHAND"] = "SecondaryHandSlot",
  ["INVTYPE_HOLDABLE"] = "SecondaryHandSlot",
  ["INVTYPE_RANGED"] = "RangedSlot",
  ["INVTYPE_THROWN"] = "RangedSlot",
  ["INVTYPE_RANGEDRIGHT"] = "RangedSlot",
  ["INVTYPE_RELIC"] = "RangedSlot",
  ["INVTYPE_TABARD"] = "TabardSlot",
}

GearEvaluator_Searcher.gearGroups = {
  ["HeadSlot"] = "armor",
  ["NeckSlot"] = "armor",
  ["ShoulderSlot"] = "armor",
  ["BackSlot"] = "armor",
  ["ChestSlot"] = "armor",
  ["ShirtSlot"] = "armor",
  ["TabardSlot"] = "armor",
  ["WristSlot"] = "armor",
  ["HandsSlot"] = "armor",
  ["WaistSlot"] = "armor",
  ["LegsSlot"] = "armor",
  ["FeetSlot"] = "armor",
  ["Finger0Slot"] = "rings",
  ["Finger1Slot"] = "rings",
  ["Trinket0Slot"] = "trinkets",
  ["Trinket1Slot"] = "trinkets",
  ["MainHandSlot"] = "weapons",
  ["SecondaryHandSlot"] = "offhands",
  -- These two can be considered as armor
  ["RangedSlot"] = "armor",
  ["AmmoSlot"] = "armor",
}

GearEvaluator_Searcher.unbound = {
  ["Binds when equipped"] = true,
  ["Binds when picked up"] = true
}

GearEvaluator_Searcher.mutexGroups = {
  ["rings"] = { "Finger0Slot", "Finger1Slot" },
  ["trinkets"] = { "Trinket0Slot", "Trinket1Slot" }
}

function GearEvaluator_Searcher:OnInitialize()
  local defaults = {
    profile = {
      interval = 0.5,
      generations = 10,
      individuals = 20,
      survivors = 1,
      itemReplaceProb = 0.1
    }
  }

  self.db = GearEvaluator.db:RegisterNamespace("Searcher", defaults)
end

function GearEvaluator_Searcher:OnEnable()
  self:RegisterEvent("PLAYER_REGEN_ENABLED", "HandleRegenEvent")
  self:RegisterEvent("PLAYER_REGEN_DISABLED", "HandleRegenEvent")
  self:RegisterEvent("PLAYER_FLAGS_CHANGED", "HandleFlagsChangedEvent")
  self.invChangeBucket = self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", 0.1, "HandleInventoryChangedEvent")
  self.bagUpdateBucket = self:RegisterBucketEvent("BAG_UPDATE", 0.1, "HandleBagUpdateEvent")

  if (not UnitAffectingCombat("player")) then
    self:StartSearching()
  end
end

function GearEvaluator_Searcher:OnDisable()
  self:StopSearching()
  self:UnregisterEvent("PLAYER_REGEN_ENABLED")
  self:UnregisterEvent("PLAYER_REGEN_DISABLED")
  self:UnregisterEvent("PLAYER_FLAGS_CHANGED")
  self:UnregisterBucket(self.invChangeBucket)
  self:UnregisterBucket(self.bagUpdateBucket)
end

function GearEvaluator_Searcher:HandleRegenEvent(event)
  if (event == "PLAYER_REGEN_ENABLED") then
    self:StartSearching()
  else
    self:StopSearching()
  end
end

function GearEvaluator_Searcher:HandleInventoryChangedEvent(units)
  if (units.player) then self:BuildItemTable() end
end

function GearEvaluator_Searcher:HandleBagUpdateEvent()
  self:BuildItemTable()
end

function GearEvaluator_Searcher:HandleFlagsChangedEvent(event, unit)
  if (unit == "player") then
    self:StopSearching()
    self:StartSearching()
  end
end

function GearEvaluator_Searcher:StartSearching()
  if (not self.searchTimer and not UnitAffectingCombat("player") and UnitIsAFK("player")) then
    self:Debug(1, "Starting to search for better gear...")
    self.searchTimer = self:ScheduleRepeatingTimer("DoSearchBurst", self.db.profile.interval)
  end
end

function GearEvaluator_Searcher:StopSearching()
  if (self.searchTimer) then
    self:CancelTimer(self.searchTimer, true)
    self.searchTimer = nil
    self:Debug(1, "Stopped searching for better gear")
  end
end

function GearEvaluator_Searcher:AddItemToItemTable(tbl, item)
  local slot = self:GetEquipLocation(item)

  if (slot) then
    self:Debug(5, item.." : "..slot)

    local group = self.gearGroups[slot]

    if (group) then
      if (not tbl[group]) then
        tbl[group] = {}
      end

      if (not tbl[group][slot]) then
        tbl[group][slot] = {}
      end

      table.insert(tbl[group][slot], item)
    end
  end
end

function GearEvaluator_Searcher:BuildItemTable()
  self:StopSearching()

  -- Start from empty gear table
  self.gearTable = {}

  -- Create scanning tooltip if not available
  -- NOTE: Handling the tooltip here make things easier
  if (not self.tooltip) then
    self.tooltip = CreateFrame("GameTooltip")
    self.tooltip:SetOwner(WorldFrame, "ANCHOR_NONE")

    local lefts, rights = {}, {}

    for i = 1, 30 do
      local left, right = self.tooltip:CreateFontString(), self.tooltip:CreateFontString()
      left:SetFontObject(GameFontNormal)
      right:SetFontObject(GameFontNormal)
      self.tooltip:AddFontStrings(left, right)
      lefts[i], rights[i] = left, right
    end

    self.tooltip.lefts, self.tooltip.rights = lefts, rights
  end

  local items = {}

  for k, slot in pairs(self.slotNames) do
    local i = GetInventoryItemsForSlot(GetInventorySlotInfo(slot))
    if (i) then
      for k, v in pairs(i) do items[k] = v end
    end
  end

  for pos, item in pairs(items) do
    local player, bank, bags, slot, bag = EquipmentManager_UnpackLocation(pos)

    if (player) then
      local itemlink

      self.tooltip:ClearLines()

      if (bags) then
        itemlink = GetContainerItemLink(bag, slot)
        self.tooltip:SetBagItem(bag, slot)
      else
        itemlink = GetInventoryItemLink("player", slot)
        self.tooltip:SetInventoryItem("player", slot)
      end

      self:AddItemToItemTable(self.gearTable, itemlink)
    end
  end

  self:StartSearching()
end

function GearEvaluator_Searcher:GetEquipLocation(item)
  if (not IsEquippableItem(item)) then return nil end

  for i = 1, self.tooltip:NumLines() do
    if (self.unbound[self.tooltip.lefts[i]:GetText()]) then
      self:Debug(2, item.." is unbound, ignoring")
      return nil
    end
  end

  local loc = self:GetEquipType(item)

  local slot = loc and self.invSlots[loc]

  return slot, loc
end

function GearEvaluator_Searcher:GetEquipType(item)
  return select(9, GetItemInfo(item))
end

function GearEvaluator_Searcher:GetMutationTable()
  if (#self.dead == 0) then
    -- No recycled table available, create new table
    self:Debug(5, "Creating new mutation table")
    return {}
  end

  return table.remove(self.dead)
end

function GearEvaluator_Searcher:CreateMutation(population)
  local individuals = #population
  local mutation = self:GetMutationTable()

  -- Armor can easily be replaced by another item
  if (self.gearTable.armor) then
    for slot, items in pairs(self.gearTable.armor) do
      if (math.random() <= self.db.profile.itemReplaceProb) then
        mutation[slot] = items[math.random(#items)]
      else
        mutation[slot] = population[math.random(individuals)][slot]
      end
    end
  end

  -- Rings and trinkets fit in 2 different slots and cannot be placed in both
  for group, slots in pairs(self.mutexGroups) do
    if (self.gearTable[group] and math.random() <= self.db.profile.itemReplaceProb) then
      local items = self.gearTable[group][slots[1]]
      local firstItem = math.random(#items)
      mutation[slots[1]] = items[firstItem]

      if (#items > 1) then
        local secondItem = math.random(#items - 1)
        if (secondItem >= firstItem) then
          mutation[slots[2]] = items[secondItem + 1]
        else
          mutation[slots[2]] = items[secondItem]
        end
      end
    else
      local parent = math.random(individuals)
      mutation[slots[1]] = population[parent][slots[1]]
      mutation[slots[2]] = population[parent][slots[2]]
    end
  end

  -- Weapons and offhands require special handling
  if ((self.gearTable.weapons or self.gearTable.offhands) and math.random() <= self.db.profile.itemReplaceProb) then
    local mainHandIndex = math.random(#self.gearTable.weapons.MainHandSlot + 1) -- +1 to allow empty mainhand slot
    mutation.MainHandSlot = self.gearTable.weapons.MainHandSlot[mainHandIndex]
    local types = self:GetAllowedOffHandTypes(mutation.MainHandSlot)
    local allowedOffHands = {}

    if (self.gearTable.weapons and self.gearTable.weapons.MainHandSlot) then
      for i, item in ipairs(self.gearTable.weapons.MainHandSlot) do
	if (types[self:GetEquipType(item)] and i ~= mainHandIndex) then
	  table.insert(allowedOffHands, item)
	end
      end
    end

    if (self.gearTable.offhands and self.gearTable.offhands.SecondaryHandSlot) then
      for i, item in ipairs(self.gearTable.offhands.SecondaryHandSlot) do
	if (types[self:GetEquipType(item)]) then
	  table.insert(allowedOffHands, item)
	end
      end
    end

    mutation.SecondaryHandSlot = allowedOffHands[math.random(#allowedOffHands + 1)] -- +1 to allow empty off-hand slot
  else
    local parent = math.random(individuals)
    mutation.MainHandSlot = population[parent].MainHandSlot
    mutation.SecondaryHandSlot = population[parent].SecondaryHandSlot
  end

  return mutation
end

function GearEvaluator_Searcher:GetAllowedOffHandTypes(item)
  -- TODO: Implement this properly
  if (item and self:GetEquipType(item) == "INVTYPE_2HWEAPON") then
    return {}
  end

  return { ["INVTYPE_HOLDABLE"] = true }
end

function GearEvaluator_Searcher:CreateMutations(population)
  while (#population < self.db.profile.individuals) do
    local mutation = self:CreateMutation(population)

    if (mutation) then
      table.insert(population, mutation)
    end
  end
end

function GearEvaluator_Searcher:SelectSurvivors(population, evalname)
  local success
  local scores = {}

  for _, v in pairs(population) do
    success, scores[v] = GearEvaluator:GetScore(evalname, v)
  end

  table.sort(population, function(a, b)
    return scores[a] < scores[b]
  end)

  while (#population > self.db.profile.survivors) do
    table.insert(self.dead, table.remove(population, 1))
  end
end

function GearEvaluator_Searcher:DoSearchBurst()
  if (not self.gearTable) then return end

  if (not self.population) then
    self.population = {}
  end

  if (not self.dead) then
    self.dead = {}
  end

  for g = 1, self.db.profile.generations do
    for k in pairs(GearEvaluator.evaluators) do
      if (not self.population[k]) then
        self.population[k] = {}
        table.insert(self.population[k], GearEvaluator:GetCurrentEquipment())
      end

      self:CreateMutations(self.population[k])
      self:SelectSurvivors(self.population[k], k)
    end
  end
end
